////////////////////////////////////////////////////////////////////// 
// KinshipEmp.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "KinshipEmp.h"
#include "MathMatrix.h"
#include "PedigreeAlleleFreq.h"
#include "OutputKin.h"
#include "PreMeta.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <vector>
#include "math.h"
#include <algorithm>

double KinshipEmp::q = 0.05;
double KinshipEmp::miss = 0.05;

void KinshipEmp::Setup(Pedigree & ped,IntArray * pPheno,int count, String & traitName,FILE * log)
{
    printf("Calculating empirical kinship matrix...\n");
    fprintf(log,"Calculating empirical kinship matrix...\n");
    //How many markers we have from the pedfile?
    allPairs.Dimension(count, count);
    allPairs.Zero();

    Vector X;
    X.Dimension(count);

    //Now decide which markers to use, MAF > 0.05, Miss genotypes rate < 0.05

    //Use only markers with 2 allels
    //outByAlleleCount.Clear();
    //for(int m=0;m<ped.markerCount;m++)
    //	if(ped.CountAlleles(m) != 2) outByAlleleCount.Push(m);	

    //loop through markers to filter by maf
    int N=0;
    for(int i=0;i<ped.markerCount;i++) 
    {
       int alleleCount = CountAlleles(ped, i);
       //if no one is genotyped, mornomorphic, or marker is not diallelic
       if(alleleCount<=1 || alleleCount>2)
	  continue;

       //calculate allele frequencies
       bool tmp = EstimateFrequencies(ped,i,FREQ_ALL); 
       Vector freq	= ped.GetMarkerInfo(i)->freq;
       double maf = 0.0;
       int minor;
       if(freq[1]<freq[2]) 
       {
	  minor = 1;
	  maf = freq[1];
       }
       else
       {
	  minor=2;
	  maf = freq[2];
       }
       if(maf<q)
	  continue;

       //calculate the missing rate
       int n =0;
       for (int f = 0; f < ped.familyCount; f++)
       {
	  for(int j=0;j<pPheno[f].Length();j++)
	  {
	     if(!ped[pPheno[f][j]].isGenotyped(i)) 
		n++;
	  }
       }
       double rate=n/count;
       if(rate>miss)
	  continue;

       N++; //useful marker count ++
       double mean=0.0;
       double variance = 0.0;
       double mean_sqr = 0.0;
       int idx =0;
       //now fill in the genotypes
       int n_=0;
       for (int f = 0; f < ped.familyCount; f++)
       {
	  for(int j=0;j<pPheno[f].Length();j++)
	  {
	     if(!ped[pPheno[f][j]].isGenotyped(i))
	     {
		X[idx] = -1.0;
	     }
	     else if(ped[pPheno[f][j]].markers[i].isHeterozygous())
	     {
		X[idx]=1.0;
		mean += X[idx];
		mean_sqr += X[idx]*X[idx];
		n_++;
	     }
	     else if(ped[pPheno[f][j]].markers[i].isHomozygous() && ped[pPheno[f][j]].markers[i].hasAllele(minor))
	     {
		X[idx]=2.0;
		mean += X[idx];
		mean_sqr += X[idx]*X[idx];
		n_++;
	     }
	     else 
	     {
		X[idx]=0.0;
		mean += X[idx];
		mean_sqr += X[idx]*X[idx];
		n_++;
	     }
	     idx++;
	  }
       }

       mean /= n_;
       mean_sqr /= n_;
       variance = mean_sqr - mean*mean;

       for(int p=0;p<count;p++)
       {
	  if(X[p]==-1.0)
	     X[p] =0.0;
	  else
	     X[p] -= mean;
       }

       for(int r=0;r<count;r++)
       {
	  for(int c=r;c<count;c++)
	  {
	     allPairs[r][c] += X[r]*X[c]/variance;
	  }
       }
    }

    for(int i=0;i<count;i++)
       for(int j=i+1;j<count;j++)
	  allPairs[j][i] = allPairs[i][j];

    allPairs.Multiply(1.0/N);
    //for(int row=0;row<5;row++) 
    //{
    //	for(int col=0;col<5;col++)
    //	{ 
    //		printf("%g ",allPairs[row][col]);
    //	}
    //	printf("\n");
    //}
    printf("Finished calculating empirical kinship matrix. \n ");
    fprintf(log,"Finished calculating empirical kinship matrix. \n ");
    if(OutputKin::outputKin) 
    {
       printf("Saving empirical kinship matrix...\n");
       fprintf(log,"Saving empirical kinship matrix...\n");
       FILE * file = NULL;
       String filename;
       if(PreMeta::outputFile != "")
       {
	  filename.printf("%s.%s.Empirical.Kinship",(const char *) PreMeta::outputFile,traitName.c_str());
       }
       else
       {
	  filename.printf("%s.Empirical.Kinship",traitName.c_str());
       }
       file = fopen(filename, "wt");
       for (int f = 0; f < ped.familyCount; f++)
       {  
	  for(int j=0;j<pPheno[f].Length();j++)
	  {  
	     fprintf(file,"%s ",ped[pPheno[f][j]].pid.c_str());
	  }
       }
       fprintf(file,"\n");

       for(int r=0;r<allPairs.rows;r++)
       {
	  for(int c=0;c<allPairs.cols;c++)
	  {
	     fprintf(file,"%g ",allPairs[r][c]);
	  }
	  fprintf(file,"\n");
       }
       fclose(file);
    }
}

void KinshipEmp::SetupVCF(Pedigree & ped, IntArray * pPheno, StringIntHash & foundersHash,StringIntHash & sampleVCFIDHash,int count,String & traitName,FILE * log)
{
   printf("Calculating empirical kinship matrix...\n");
   fprintf(log,"Calculating empirical kinship matrix...\n");
   VcfFileReader reader;
   VcfHeader header;
   VcfRecord record;
   reader.open(PreMeta::vcfInput,header);

   allPairs.Dimension(count,count);
   allPairs.Zero();

   Vector genotype;
   genotype.Dimension(count);
   genotype.Zero();
   int N=0;
   if(PreMeta::dosage)
   {
      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 int n=0,nf=0,nmiss=0;
	 double fmaf=0.0,maf=0.0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<pPheno[f].Length();j++)
	    {
	       int s = sampleVCFIDHash.Integer(ped[pPheno[f][j]].pid);
	       const std::string * geno = genoInfo.getString("DS",s);

	       if(*geno == ".")
	       {
		  nmiss++;
		  continue;
	       }
	       float dose = atof((*geno).c_str());

	       //judge if this is a founder
	       int founder = foundersHash.Integer(ped[pPheno[f][j]].pid);
	       if(founder != -1)
	       {
		  fmaf += dose;
		  nf++;
	       }
	       //Calculate overall maf
	       maf += dose;
	       n++;
	    }
	 }
	 //calculate final maf
	 if(nf!=0)
	    fmaf /= 2.0*nf;
	 maf /= 2.0*n;
	 if(fmaf==0.0 && maf >0.0)
	    fmaf=maf;

	 if(PreMeta::FounderFreq)
	 {
	    if(nmiss > count*miss || fmaf<q)
	    {
	       continue;
	    }
	 }
	 else
	 {
	    if(nmiss > count*miss || maf<q)
	    {
	       continue;
	    }
	 }
	 N++;

	 //step2: if yes, then loop through all samples to get genotype. 
	 //genotype.Clear();
	 double mean = 0.0, mean_sqr= 0.0;
	 int n_ = 0;
	 int idx =0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<pPheno[f].Length();j++)
	    {
	       int s = sampleVCFIDHash.Integer(ped[pPheno[f][j]].pid);
	       //fill in genotype vector for this sample
	       const std::string * geno = genoInfo.getString("DS",s);
	       if(*geno == ".")
		  genotype[idx] = _NAN_;
	       else
	       {
		  genotype[idx] = atof((*geno).c_str());
		  mean += genotype[idx];
		  n_++;
		  mean_sqr += genotype[idx]*genotype[idx];
	       }
	       idx++;
	    }
	 }

	 //step 3: standardize genotype vector
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    else
	       genotype[i] -= mean;
	 }
	 //add up to kinship
	 for(int r=0;r<count;r++)
	 {
	    for(int c=r;c<count;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	    }
	 }
      }

      for(int i=0;i<count;i++)
	 for(int j=i+1;j<count;j++)
	    allPairs[j][i] = allPairs[i][j];

      allPairs.Multiply(1.0/N);
   }
   else
   {
      IntArray counts;
      counts.Dimension(2);

      while(reader.readRecord(record))
      {
	 //step1: decide if this marker should be included
	 double maf=0.0;
	 int nmiss =0;
	 counts.Zero();
	 //	 fcounts.Zero();
	 bool skip = false;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int p=0;p<pPheno[f].Length();p++)
	    {
	       bool miss_stat = false;
	       int i = sampleVCFIDHash.Integer(ped[pPheno[f][p]].pid);
	       int numGTs = record.getNumGTs(i);
	       for(int j = 0; j < numGTs; j++)
	       {
		  int a = record.getGT(i,j);
		  if(a==VcfGenotypeSample::MISSING_GT)
		  {
		     miss_stat=true;
		     nmiss++;
		     break;
		  }
		  if(a>1)
		  {
		     skip = true;
		     //if marker is not biallelic, skip this marker
		     goto skip_to_point;
		  }
		  counts[a]++;
	       }
	       if(miss_stat)
	       {
		  //if genotype is missing, skip this sample.
		  continue;
	       }

	       //judge if this is a founder
	       //	       int founder = foundersHash.Integer(ped[pPheno[f][p]].pid);
	       //		  if(founder != -1)
	       //		  {
	       //		     if(allele[m] >= fcounts.size())
	       //		     {
	       //			fcounts.resize(allele[m]+1);
	       //		     }
	       //		     ++fcounts[allele[m]];
	       //		}
	    }
	 }

skip_to_point: 
	 if(skip)
	    continue;

	 if(counts[0]==0 || counts[1]==0)
	    continue;

	 //Calculate maf
	 maf = (double)counts[1]/(counts[0]+counts[1]);
	 if(maf>0.5) maf = 1.0-maf;

	 if(nmiss > count*miss || maf<q)
	 {
	    continue;
	 }

	 N++;
	 //printf("%d\n",N);

	 //step2: if yes, then loop through all samples to get genotype. 

	 int n_=0;
	 double mean = 0.0,mean_sqr=0.0;
	 int idx =0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int p=0;p<pPheno[f].Length();p++)
	    {
	       int s = sampleVCFIDHash.Integer(ped[pPheno[f][p]].pid);
	       int numGTs = record.getNumGTs(s);
	       bool skip=false;
	       for(int j = 0; j < numGTs; j++)
	       {
		  int a = record.getGT(s,j);
		  if(a==VcfGenotypeSample::MISSING_GT)
		  {
		     genotype[idx] = _NAN_;
		     skip=true;
		     break;
		  }
		  genotype[idx] += a;
	       }
	       if(skip)
		  continue;

	       mean += genotype[idx];
	       n_++;
	       mean_sqr += genotype[idx]*genotype[idx];
	       idx++;
	    }
	 }

	 //step 3: standardize genotype vector
	 mean /= n_;
	 double variance = mean_sqr/n_ - mean*mean;
	 if(record.get1BasedPosition()==10870838)
	    printf("chr%s %d %d %g %g\n",record.getChromStr(),record.get1BasedPosition(),N,mean,variance);
	 for(int i=0;i<genotype.Length();i++)
	 {
	    if(genotype[i]==_NAN_)
	    {
	       //make sure if genotype is missing, then it is not contributing
	       genotype[i]=0.0;
	    }
	    else
	    {
	       genotype[i] -= mean;
	    }
	 }
	 //add up to kinship
	 //	 printf("%s %s variance =%g\n",record.getChromStr(),record.get1BasedPosition(),variance);
	 for(int r=0;r<count;r++)
	 {
	    for(int c=r;c<count;c++)
	    {
	       allPairs[r][c] += genotype[r]*genotype[c]/variance;
	       //  printf("%g ",allPairs[r][c]);
	    }
	    //printf("\n");
	 }
      }

      for(int i=0;i<count;i++)
	 for(int j=i+1;j<count;j++)
	    allPairs[j][i] = allPairs[i][j];

      if(N==0)
	 error("ERROR! No qualified marker to use for kinship calculation.\n");

      allPairs.Multiply(1.0/N);
   }

   if(OutputKin::outputKin)
   {
      printf("Writing kinship matrix ...\n");
      fprintf(log,"Writing kinship matrix ...\n");
      FILE * file = NULL;
      String filename;
      if(PreMeta::outputFile != "")
      {
	 filename.printf("%s.%s.Empirical.Kinship",(const char *) PreMeta::outputFile,traitName.c_str());
      }
      else
      {
	 filename.printf("%s.Empirical.Kinship",traitName.c_str());
      }
      file = fopen(filename, "wt");
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<pPheno[f].Length();j++)
	 {
	    fprintf(file,"%s ",ped[pPheno[f][j]].pid.c_str());
	 }
      }
      fprintf(file,"\n");
      for(int r=0;r<allPairs.rows;r++)
      {
	 for(int c=0;c<allPairs.cols;c++)
	 {
	    fprintf(file,"%g ",allPairs[r][c]);
	 }
	 fprintf(file,"\n");
      }
      fclose(file);
   }
}


